/****************************************************************************
*
*  stripwav.cpp - strip leading and trailing silence from wave file
*
*  Release: 0.1
*  Create:  24.01.2000
*  Update:  03.10.2001
*  (C)      Marcel Mller
*
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <ctype.h>
#define INCL_DOSFILEMGR
#define INCL_DOSPROCESS
#include <os2.h>

#define n (1*1024*1024)
#define maxscan 64
#define minstrip 10000

// bad game ...
#undef min
#undef max
template <class X>
inline const X min(const X &a, const X &b)
{  return (a>b)?b:a;
}
template <class X>
inline const X max(const X &a, const X &b)
{  return (a<b)?b:a;
}

static HFILE f;
static int thres1 = -1;
static int thres2 = -1;
static int trunc = 0;
static char* fname = NULL;
static short b[n/2];
static unsigned int s_front[maxscan+1];
static unsigned int s_back[maxscan+1];

static void error(unsigned long r, const char* fmt, int ret = 1)
{  DosClose(f);
   fprintf(stderr, fmt, r);
   exit(1);
}
static inline void errchk(unsigned long r, const char* fmt, int ret = 1)
{  if (r)
      error(r, fmt, ret);
}

template <class T>
static inline T* ptradd(T* p1, unsigned long d)
{  return (T*)((char*)p1 + d);
}

template <class T>
static inline long ptrdif(T* p1, T* p2)
{  return ((char*)p1 - (char*)p2);
}


static void seekwav(long offset)
{  unsigned long r2;
   errchk(DosSetFilePtr(f, offset, FILE_CURRENT, &r2), "Failed to set file pointer - Error %lu. #1\n");
}

static void setwav(long offset)
{  unsigned long r2;
   errchk(DosSetFilePtr(f, offset, FILE_BEGIN, &r2), "Failed to set file pointer - Error %lu. #2\n");
}

static unsigned long readwav(unsigned long l)
{  if (l > n)
      l = n;
   unsigned long r2;
   unsigned long r = DosRead(f, b, l, &r2);
   if (r || r2 != l)
      error(r, "Failed to read wave data - Error %lu. #1\n");
   return l;
}

static unsigned long rreadwav(unsigned long l, unsigned long r2)
{  if (l > n)
      l = n;
   seekwav(-l-r2);
   unsigned long r = DosRead(f, b, l, &r2);
   if (r || r2 != l)
      error(r, "Failed to read wave data - Error %lu. #2\n");
   return l;
}

static void writewav(void* buf, unsigned long l)
{  unsigned long r2;
   errchk(DosWrite(f, buf, l, &r2), "Failed to write wave data - Error %lu.\n");
   if (r2 != l)
      error(0, "Failed to write wave data - Disk full.\n");
}

static int autoscan(unsigned long off, unsigned long l)
{  register short* sp = b;
   register short* end = ptradd(b, l &= ~3);
   for (; sp < end; sp += 2)
   {  short val;
      if ((val = max(abs(sp[0]),abs(sp[1]))) <= thres1)
         continue;
      //printf("%i %i\n", val, ptrdif(sp, b) + off);
      if (val >= maxscan)
      {  s_front[maxscan] = ptrdif(sp, b) + off;
         return 1;
      }
      s_front[val] = ptrdif(sp, b) + off;
      thres1 = val;
   }
   return 0;
}

static int rautoscan(unsigned long off, unsigned long l)
{  register short* sp = ptradd(b, l -4);
   register short* end = sp;
   for (; sp >= b; sp -= 2)
   {  short val;
      if ((val = max(abs(sp[0]),abs(sp[1]))) <= thres2)
         continue;
      if (val >= maxscan)
      {  s_back[maxscan] = ptrdif(end, sp) + off;
         return 1;
      }
      s_back[val] = ptrdif(end, sp) + off;
      thres2 = val;
   }
   return 0;
}

static unsigned long scanbuf(unsigned long len)
{  register short* sp = b;
   unsigned long l = len;
   while (l && abs(*sp++) <= thres1)
      --l;
   return (len-l) & ~1;
}

static unsigned long rscanbuf(unsigned long len)
{  register short* sp = b + len;
   unsigned long l = len;
   while (l && abs(*--sp) <= thres2)
      --l;
   return (len-l) & ~1;
}

int main(int argc, char** argv)
{  if (argc < 2)
   {  fputs("Usage: STRIPWAVE <filename> [<threshold> [<threshold at end>]] [<options>]\n"
            "If threshold is omitted you will be asked for it after a scan.\n"
            "/t  Truncate additional data.\n", stderr);
      return 6;
   }
   while (--argc)
   {  char* cp = *++argv;
      if (*cp == '/')
      {  // option
         switch (toupper(cp[1]))
         {default:
            fprintf(stderr, "Illegal option %s.\n", cp);
            return 5;
          case 'T':
            trunc = 1;
         }
      } else if (!fname)
         fname = cp;
       else if (thres1 == -1)
      {  int r = 0;
         sscanf(cp, "%i%n", &thres1, &r);
         if (!r || thres1 < 0)
         {  fprintf(stderr, "Threshold %s is not a nonnegative integer.\n", cp);
            return 5;
         }
      } else if (thres2 == -1)
      {  int r = 0;
         sscanf(cp, "%i%n", &thres2, &r);
         if (!r || thres2 < 0)
         {  fprintf(stderr, "Threshold %s is not a nonnegative integer.\n", cp);
            return 5;
         }
      } else
      {  fprintf(stderr, "Unexpected argument %s.\n", cp);
         return 5;
      }
   }
   if (!fname)
   {  fputs("No filename given.\n", stderr);
      return 5;
   }

   ULONG r, r2;
   r = DosOpen((PSZ)fname, &f, &r2, 0, 0, OPEN_ACTION_FAIL_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS, OPEN_FLAGS_RANDOM|OPEN_SHARE_DENYWRITE|OPEN_ACCESS_READWRITE, NULL);
   if (r)
   {  fprintf(stderr, "Cannot open %s - Error %lu.\n", fname, r);
      return 4;
   }

   struct
   {  char           RIFF[4];
      unsigned long  BlkLen;
      char           WAVE[8];
      unsigned long  WAVLen;
      unsigned short x1;
      unsigned short Channels;
      unsigned long  SampRate;
      unsigned long  x2;
      unsigned short BpSamp;
      unsigned short Bits;
      char           data[4];
      unsigned long  DataLen;
   } Header;

   errchk(DosRead(f, &Header, sizeof Header, &r2), "Failed to read header - Error %lu.\n", 3);
   if (r2 != sizeof Header)
      error(r2, "Failed to read header - Unexpected end of file after %lu bytes.\n", 3);

   if (  memcmp(Header.RIFF, "RIFF", sizeof Header.RIFF)
      || memcmp(Header.WAVE, "WAVEfmt ", sizeof Header.WAVE)
      || Header.WAVLen != 16
      || Header.BpSamp != 4
      || Header.Bits != 16
      || memcmp(Header.data, "data", sizeof Header.data) )
   {  DosClose(f);
      fputs("Unrecognized file format.", stderr);
      return 2;
   }

   FILESTATUS3 fst;
   errchk(DosQueryFileInfo(f, FIL_STANDARD, &fst, sizeof fst), "Failed to query file info - Error %lu.\n");

   if (Header.BlkLen + 8 != fst.cbFile || Header.DataLen + sizeof Header > fst.cbFile)
   {  DosClose(f);
      fprintf(stderr, "Wave data size does not match - Block: %lu, Data: %lu, File: %lu.\n", Header.BlkLen, Header.DataLen, fst.cbFile);
      return 2;
   }

   unsigned long len = Header.DataLen;
   unsigned long offset = 0;
   unsigned long l, ls;

   if (thres1 == -1)
   {  // perform autoscan
      fputs("Analyzing data...\n", stderr);
      // scan front
      memset(s_front, -1, sizeof s_front);
      ls = len;
      thres1 = 0;
      for (;;)
      {  l = readwav(ls);
         if (autoscan(offset, l) || (ls -= l) == 0)
            break;
         offset += l;
      }
      offset = 0;

      setwav(Header.DataLen + sizeof Header);
      // scan back
      memset(s_back, -1, sizeof s_back);
      ls = len;
      thres2 = 0;
      l = 0;
      for (;;)
      {  l = rreadwav(ls, l);
         if (rautoscan(offset, l) || (ls -= l) == 0)
            break;
         offset += l;
      }
      offset = 0;

      // prepare data
      unsigned int* fp = s_front + maxscan +1;
      unsigned int* bp = s_back + maxscan +1;
      l = len;
      ls = len;
      r = maxscan;
      do
      {  if (*--fp != (unsigned)-1)
            l = *fp;
          else
            *fp = l;
         if (*--bp != (unsigned)-1)
            ls = *bp;
          else
            *bp = ls;
      } while (r--);
      *fp = 0;
      *bp = 0;

      // display results
      fputs("Level\tFront\tF-Diff.\tBack\tB-Diff.\tSum\tDiff.\n", stderr);
      fp = s_front + maxscan +1;
      bp = s_back + maxscan +1;
      l = *--fp;
      ls = *--bp;
      unsigned int df, db;
      unsigned int mdf = minstrip;
      unsigned int mdb = minstrip;
      thres1 = -1;
      thres2 = -1;
      for (r = maxscan; r--;)
      {  if (*--fp + *--bp == l + ls)
            continue;
         fprintf(stderr, "T: %lu\t", r);
         if (*fp != l)
         {  df = l - *fp;
            fprintf(stderr, "%lu\t%u\t", l, df);
            if (df > mdf)
            {  mdf = df;
               thres1 = r;
            }
         } else
            fputs("\t\t", stderr);
         if (*bp != ls)
         {  db = ls - *bp;
            fprintf(stderr, "%lu\t%u\t", ls, db);
            if (db > mdb)
            {  mdb = db;
               thres2 = r;
            }
         } else
            fputs("\t\t", stderr);
         fprintf(stderr, "%lu\t%u\n", l+ls, df+db);
         l = *fp;
         ls = *bp;
      }

      // ask for thres
      fprintf(stderr, "Enter Threshold (auto = %i,%i):\n", thres1, thres2);
      do
      {  while (!gets((char*)b));
         mdf = (unsigned)-1;
         mdb = (unsigned)-1;
         sscanf((char*)b, "%n%u%n%*[\t ,]%u%n", &l, &mdf, &l, &mdb, &l);
      } while (l != strlen((char*)b));
      if (mdf != (unsigned)-1)
      {  thres1 = mdf;
         thres2 = mdb == (unsigned)-1 ? mdf : mdb;
      }
      //printf("%u %u %u %u\n", mdf, mdb, thres1, thres2);

      // reset file ptr
      setwav(sizeof Header);

   } else if (thres2 == -1)
      thres2 = thres1;

   // strip front
   while (len)
   {  l = readwav(len);
      len -= l;
      l >>= 1;
      offset += ls = scanbuf(l);
      if (ls < l)
         break;
   }
   offset <<= 1;
   printf("Removed %lu bytes at the front.\n", offset);

   if (offset)
   {  // move data
      setwav(sizeof Header);
      l = (l-ls)<<1;
      writewav(b+ls, l);
      while (len)
      {  fprintf(stderr, "%lu bytes remaining to move.   \r", len);
         fflush(stderr);
         seekwav(offset);
         l = readwav(len);
         len -= l;
         seekwav(-l-offset);
         writewav(b, l);
      }
   } else
      setwav(Header.DataLen + sizeof Header);

   // strip back
   len = Header.DataLen - offset;
   unsigned long offset2 = 0;
   l = 0;
   while (len)
   {  l = rreadwav(len, l);
      len -= l;
      unsigned long ln = l >> 1;
      offset2 += ls = rscanbuf(ln);
      if (ls < ln)
         break;
   }
   offset += offset2 <<= 1;
   printf("Removed %lu bytes at the end.      \n", offset2);

   // move any remaining data
   len = fst.cbFile - sizeof Header - Header.DataLen;
   if (!trunc)
   {  while (len)
      {  fprintf(stderr, "%lu bytes remaining to move.   \n", len);
         fflush(stderr);
         setwav(fst.cbFile - len);
         l = readwav(len);
         len -= l;
         seekwav(-l-offset);
         writewav(b, l);
      }
      len = 0;
   }
   len += offset;
   if (len)
   {  // truncate file
      errchk(DosSetFileSize(f, fst.cbFile - len), "Failed to truncate file - Error %lu.\n");

      // update wave header
      Header.BlkLen -= len;
      Header.DataLen -= offset;

      // write wave header
      setwav(0);
      writewav(&Header, sizeof Header);
   }

   DosClose(f);

   // time to see the result
   DosSleep(1000);
   return 0;
}
